from pathlib import Path

import lmfit
import matplotlib.pyplot as plt
import mrcfile
import numpy as np
from matplotlib import patches
from scipy import ndimage
from skimage import filters
from skimage.registration import phase_cross_correlation
from tqdm.auto import tqdm

from .utils import bin_ndarray
from .utils import translation


DEFAULTS = {
    'direction': 'downup',
    'zigzag': True,
    'flip': True,
    'rot90': 3,
    'flipud': True,
    'fliplr': False,
}


def sorted_grid_indices(grid):
    """Sorts 2d the grid by its values, and returns an array with the indices
    (i.e. np.argsort on 2d arrays) https://stackoverflow.com/a/30577520."""
    return np.dstack(np.unravel_index(np.argsort(grid.ravel()), grid.shape))[0]


def find_threshold(values, half='lower') -> float:
    """Find largest discontinuity in the `lower` or `upper` half of the data.

    Parameters
    ----------
    half : str
        Whether to use the `upper` or `lower` half of the data after sorting

    Returns
    -------
    threshold : float
        Threshold splitting the largest discontinuity to segment data with
    """
    x = np.array(sorted(values))
    halfway = int(len(x) / 2)
    sel = x[:halfway] if half == 'lower' else x[halfway:]
    diff = np.diff(sel)
    i = diff.argmax()
    if half == 'upper':
        i += halfway
    thresh = (x[i] + x[i + 1]) / 2
    return thresh


def weight_map(shape, method='block', plot=False):
    """Generate a weighting map for the given shape.

    Parameters
    ----------
    shape : tuple
        Shape defines the 2 integers defining the shape of the image
    method : str
        Method to use `circle`/`block`
    plot : bool
        Plot the image

    Returns
    -------
    weight : np.array
        Weight array with the given shape
    """
    res_x, res_y = shape
    c_x = int(res_x / 2) - 0.5
    c_y = int(res_y / 2) - 0.5

    corner = (c_x**2 + c_y**2)**0.5

    a, b = np.meshgrid(np.arange(-c_x, c_x + 1), np.arange(-c_y, c_y + 1))

    if method == 'block':
        a2 = c_x - np.abs(a)
        b2 = c_y - np.abs(b)

        d = np.min(np.stack((a2, b2)), axis=0)
    elif method == 'circle':
        d = corner - np.sqrt(a**2 + b**2)
    else:
        raise ValueError(f'No such method: `{method}`')

    # scale to 1
    d = d / d.max()

    if plot:
        plt.imshow(d)

    return d


def make_grid(gridshape: tuple,
              direction: str = 'updown',
              zigzag: bool = True,
              flip: bool = False,
              ) -> 'np.array':
    """Defines the grid montage collection scheme.

    Parameters
    ----------
    gridshape : tuple(int, int)
        Defines the shape of the grid
    direction : str
        Defines the direction of data collection starting from the top (lr, rl)
        or left-hand (ud, du) side `updown`, `downup`, `leftright`, `rightleft`
    zigzag : bool
        Defines if the data has been collected in a zigzag manner
    flip : bool
        Flip around the vertical (lr, rl) or horizontal (ud, du) axis, i.e. start from the
        botton (lr, rl) or right-hand (ud, du) side.

    Returns
    -------
    np.array
    """
    nx, ny = gridshape
    grid = np.arange(nx * ny).reshape(gridshape)

    if zigzag:
        grid[1::2] = np.fliplr(grid[1::2])

    if direction == 'updown':
        if flip:
            grid = np.fliplr(grid)
    elif direction == 'downup':
        grid = np.flipud(grid)
        if flip:
            grid = np.fliplr(grid)
    elif direction == 'rightleft':
        grid = grid.T
        grid = np.fliplr(grid)
        if flip:
            grid = np.flipud(grid)
    elif direction == 'leftright':
        grid = grid.T
        if flip:
            grid = np.flipud(grid)
    else:
        raise ValueError(f'Invalid direction: {direction}')

    return grid


def make_slices(overlap_x: int, overlap_y: int, shape=(512, 512), plot: bool = False) -> dict:
    """Make slices for left/right/top/bottom image.

    Parameters
    ----------
    overlap_x/overlap_y : int
        Defines how far to set the overlap from the edge (number of pixels),
        this corresponds to the overlap between images
    shape : tuple:
        Define the shape of the image (only for plotting)
    plot : bool
        Plot the boundaries on blank images

    Returns
    -------
    Dictionary with the slices for each side
    """
    d = {}

    s_right = np.s_[:, -overlap_x:]
    s_left = np.s_[:, :overlap_x]
    s_top = np.s_[:overlap_y]
    s_bottom = np.s_[-overlap_y:]

    slices = (s_right, s_left, s_top, s_bottom)
    labels = ('right', 'left', 'top', 'bottom')

    d = dict(zip(labels, slices))

    if plot:
        fig, axes = plt.subplots(2, 2, sharex=True, sharey=True)
        axes = axes.flatten()

        for ax, s_, label in zip(axes, slices, labels):
            arr = np.zeros(shape, dtype=int)
            arr[s_] = 1
            ax.imshow(arr)
            ax.set_title(label)

        plt.show()

    return d


def define_directions(pairs: list):
    """Define pairwise relations between indices.

    Takes a list of index pair dicts, and determines on which side they
    are overlapping. The dictionary is updated with the keywords
    `side0`/`side1`.
    """
    for pair in pairs:
        i0, j0 = pair['idx0']
        i1, j1 = pair['idx1']

        # checked 21-11-2019 for 'leftright' config
        if j0 == j1:
            if i1 > i0:
                side0, side1 = 'bottom', 'top'
            else:
                side0, side1 = 'top', 'bottom'
        else:
            if j1 > j0:
                side0, side1 = 'right', 'left'
            else:
                side0, side1 = 'left', 'right'

        # print(i0, j0, i1, j1, side0, side1)

        pair['side0'] = side0
        pair['side1'] = side1

    return pairs


def define_pairs(grid: 'np.ndarray'):
    """Take a sequence grid and return all pairs of neighbours.

    Returns a list of dictionaries containing the indices of the pairs
    (neighbouring only), and the corresponding sequence numbers
    (corresponding to the image array)
    """
    nx, ny = grid.shape

    footprint = np.array([[0, 1, 0],
                          [1, 0, 1],
                          [0, 1, 0]])

    shape = np.array(footprint.shape)
    assert shape[0] == shape[1], 'Axes must be equal'
    assert shape[0] % 2 == 1, 'Axis length must be odd'
    center = shape // 2

    connected = np.argwhere(footprint == 1) - center

    pairs = []

    for idx0, i0 in np.ndenumerate(grid):
        neighbours = connected + idx0

        for neighbour in neighbours:
            neighbour = tuple(neighbour)
            if neighbour[0] < 0 or neighbour[0] >= nx:
                pass
            elif neighbour[1] < 0 or neighbour[1] >= ny:
                pass
            else:
                assert i0 == grid[idx0]
                d = {
                    'seq0': grid[idx0],
                    'seq1': grid[neighbour],
                    'idx0': idx0,
                    'idx1': neighbour,
                }
                pairs.append(d)

    return pairs


def disambiguate_shift(strip0, strip1, shift, verbose: bool = False) -> tuple:
    """Disambiguate the shifts obtained from cross correlation."""
    shift_x, shift_y = shift

    best_sum = np.inf
    best_shift = shift

    for i in (-1, 1):
        for j in (-1, 1):
            new_shift = (i * shift_x, j * shift_y)
            strip1_offset = ndimage.shift(strip1, new_shift)
            offset = strip1_offset - strip0.astype(float)
            sum_score = np.abs(offset).sum()
            if verbose:
                print(f'{i:2d} {j:2d} -> {sum_score:10.0f}  {new_shift}')
            if sum_score < best_sum:
                best_sum = sum_score
                best_shift = new_shift

    if verbose:
        print('Disambiguated shift:', best_shift)

    return best_shift


def plot_images(im0, im1, seq0, seq1, side0, side1, idx0, idx1):
    fig, axes = plt.subplots(ncols=2, figsize=(6, 3))
    ax0, ax1 = axes.flatten()

    ax0.imshow(im0)
    ax0.set_title(f'{seq0} {idx0} {side0}')
    ax0.set_axis_off()
    ax1.imshow(im1)
    ax1.set_title(f'{seq1} {idx1} {side1}')
    ax1.set_axis_off()
    plt.tight_layout()
    plt.show()


def plot_fft(strip0, strip1, shift, fft, side0, side1):
    fig, axes = plt.subplots(nrows=4, figsize=(8, 5))
    axes = axes.flatten()
    for ax in axes:
        ax.set_axis_off()
    ax0, ax1, ax2, ax3 = axes

    assert strip0.shape == strip1.shape, f'Shapes do not match, strip1: {strip1.shape} strip2: {strip2.shape}'
    shape = strip0.shape

    if shape[0] > shape[1]:
        strip0 = strip0.T
        strip1 = strip1.T
        fft = fft.T
        t1, t0 = shift
    else:
        t0, t1 = shift

    # Show difference
    strip1_shifted = ndimage.shift(strip1, (t0, t1))
    difference = strip1_shifted - strip0.astype(float)

    ax0.imshow(strip0, interpolation='nearest')
    ax0.set_title(f'{side0}')
    ax1.imshow(strip1, interpolation='nearest')
    ax1.set_title(f'{side1}')
    ax2.imshow(difference, interpolation='nearest')
    ax2.set_title(f'Abs(Difference) - Shift: {t0} {t1}')
    ax3.imshow(fft, vmin=np.percentile(fft, 90.0), vmax=np.percentile(fft, 99.99))
    ax3.set_title(f'Cross correlation (max={fft.max():.4f})')

    if t0 < 0:
        t0 += shape[0]
    if t1 < 0:
        t1 += shape[0]

    ax3.scatter(t1, t0, color='red', marker='o', facecolor='none', s=100)
    ax3.set_xlim(0, fft.shape[1])
    ax3.set_ylim(0, fft.shape[0])

    plt.subplots_adjust(hspace=0.0)
    plt.show()


def plot_shifted(im0, im1, difference_vector, seq0, seq1, idx0, idx1, res_x, res_y):
    blank = np.zeros((res_x * 2, res_y * 2), dtype=np.int32)

    center = np.array(blank.shape) // 2
    origin = np.array((res_x, res_y)) // 2

    coord0 = (center - difference_vector / 2 - origin).astype(int)
    coord1 = (center + difference_vector / 2 - origin).astype(int)

    print(f'Coord0: {coord0} | Coord1: {coord1}')

    txt = f'Difference vector\n#{seq0}:{idx0} -> #{seq1}:{idx1} = {difference_vector}'

    blank[coord0[0]: coord0[0] + res_x, coord0[1]: coord0[1] + res_y] += im0
    blank[coord1[0]: coord1[0] + res_x, coord1[1]: coord1[1] + res_y] += im1

    # Create a Rectangle patch
    rect0 = patches.Rectangle(coord0[::-1], res_x, res_y, linewidth=1, edgecolor='r', facecolor='none')
    rect1 = patches.Rectangle(coord1[::-1], res_x, res_y, linewidth=1, edgecolor='r', facecolor='none')

    fig, ax = plt.subplots(1, figsize=(8, 8))

    # Add the patch to the Axes
    ax.add_patch(rect0)
    ax.add_patch(rect1)

    ax.imshow(blank)
    ax.set_title(txt)
    ax.set_axis_off()
    plt.show()


class MontagePatch:
    """Simple class to calculate the bounding box for an image tile as part of
    a montage.

    Parameters
    ----------
    image : np.ndarray [ m x n ]
        Original image tile
    coord : tuple
        Tuple with x/y coordinates of the patch (original size) location in the montage image
    binning : int
        Binning of the image patch
    """

    def __init__(self, image, coord, binning: int = 1):
        super().__init__()
        self.binning = binning
        self.coord = coord
        self._image = image
        self._shape = image.shape

    @property
    def shape(self):
        res_x, res_y = self._shape
        shape = int(res_x / self.binning), int(res_y / self.binning)
        return shape

    @property
    def res_x(self):
        return self.shape[0]

    @property
    def res_y(self):
        return self.shape[1]

    @property
    def image(self):
        return bin_ndarray(self._image, self.shape)

    @property
    def x0(self):
        return int(self.coord[0] / self.binning)

    @property
    def x1(self):
        return self.x0 + self.shape[0]

    @property
    def y0(self):
        return int(self.coord[1] / self.binning)

    @property
    def y1(self):
        return self.y0 + self.shape[1]


class Montage:
    """This class is used to stitch together a set of images to make a larger
    image.

    Parameters
    ----------
    images : list
        List of images in numpy format
    gridspec : dict
        Dictionary defining the grid characteristics, directly passed to `make_grid`.
    overlap : float
        Defines the % of overlap between the images

    Based on Preibisch et al. (2009), Bioinformatics, 25(11):1463-1465
             http://dx.doi.org/10.1093/bioinformatics/btp184
    """

    def __init__(self,
                 images: list,
                 gridspec: dict,
                 overlap=0.1,
                 **kwargs,
                 ):
        super().__init__()

        self.images = images
        self.image_shape = images[0].shape
        self.gridspec = gridspec
        self.grid = make_grid(**gridspec)

        self.__dict__.update(**kwargs)

        res_x, res_y = self.image_shape
        self.overlap_x = int(res_x * overlap)
        self.overlap_y = int(res_y * overlap)

    @classmethod
    def from_serialem_mrc(cls,
                          filename: str,
                          gridshape: tuple,
                          direction: str = DEFAULTS['direction'],
                          zigzag: bool = DEFAULTS['zigzag'],
                          flip: bool = DEFAULTS['flip'],
                          image_rot90: int = DEFAULTS['rot90'],
                          image_flipud: bool = DEFAULTS['flipud'],
                          image_fliplr: bool = DEFAULTS['fliplr'],
                          ):
        """Load a montage object from a SerialEM file image stack. The default
        parameters transform the images to data suitable for use with
        `pyserialem.Montage`. It makes no further assumptions about the way the
        data were collected.

        The parameters image_rot90/image_flipud/image_fliplr manipulate the images in this order,
        so they can look the same as when collected with Instamatic. This is necessary to use
        the stage calibration, which is not specified in the SerialEM mrc file.

        Use .set_stagematrix and .set_pixelsize to set the correct stagematrix and pixelsize.

        Parameters
        ----------
        filename : str
            Filename of the mrc file to load.
        gridshape : tuple(2)
            Tuple describing the number of of x and y points in the montage grid.
            TODO: Find a way to get this from the .mrc/.mdoc
        direction : str
            Defines the direction of data collection
        zigzag : bool
            Defines if the data has been collected in a zigzag manner
        flip : bool
            Flip around the vertical (lr, rl) or horizontal (ud, du) axis,
            i.e. start from the botton (lr, rl) or right-hand (ud, du) side.
        image_rot90 : int
            Rotate the image by 90 degrees (clockwise) for this times, i.e.,
            `image_rot90=3` rotates the image by 270 degrees.
        image_flipud:
            Flip the images around the horizintal axis.
        image_fliplr:
            Flip the images around the vertical axis. The


        Returns
        -------
        Montage object constructed from the given images
        """
        from pyserialem import read_mdoc_file
        filename = str(filename)  # in case of Path object

        gm = mrcfile.open(filename)
        images = gm.data

        mdoc = read_mdoc_file(filename + '.mdoc', only_kind='zvalue')
        assert len(mdoc) == len(images)

        gridspec = {
            'gridshape': gridshape,
            'direction': direction,
            'zigzag': zigzag,
            'flip': flip,
        }

        # Rotate the images so they are in the same orientation as those from Instamatic
        if image_rot90:
            images = [np.rot90(image, k=image_rot90) for image in images]
        if image_flipud:
            images = [np.flipud(image) for image in images]
        if image_fliplr:
            images = [np.fliplr(image) for image in images]

        kwargs = {
            'stagecoords': np.array([d['StagePosition'] for d in mdoc]) * 1000,  # um->nm
            'magnifiation': mdoc[-1]['Magnification'],
            'image_binning': mdoc[-1]['Binning'],
            'software': 'serialem',
            'abs_mag_index': mdoc[-1]['MagIndex'],
            'mdoc': mdoc,
            'filename': filename,
        }

        m = cls(images=images, gridspec=gridspec, **kwargs)

        c1 = np.array([d['PieceCoordinates'][0:2] for d in mdoc])

        def convert_coords(c,
                           rot90=image_rot90,
                           flipud=image_flipud,
                           fliplr=image_fliplr,
                           ):
            # SerialEM uses a different convention for X/Y
            c = np.fliplr(c)

            angle = np.radians(image_rot90 * 90)
            R = np.array([np.cos(angle), -np.sin(angle), np.sin(angle), np.cos(angle)]).reshape(2, 2)

            c = np.dot(c, R)

            if flipud:
                c[:, 1] *= -1
            if fliplr:
                c[:, 0] *= -1

            c -= c.min(axis=0)

            return c

        # Apparently, SerialEM can save one or the other or both
        # prefer APCVS over APC and move on
        for key in 'AlignedPieceCoordsVS', 'AlignedPieceCoords':
            if key in mdoc[0]:
                c2 = np.array([d[key][0:2] for d in mdoc])
                break

        m.coords = convert_coords(c1)
        m.optimized_coords = convert_coords(c2)

        return m

    def update_gridspec(self, **gridspec):
        """Update the grid specification."""
        self.gridspec.update(gridspec)
        self.grid = make_grid(**self.gridspec)

    def set_pixelsize(self, pixelsize: float):
        """Set the unbinned pixelsize in nanometer/pixel."""
        self.pixelsize = pixelsize * self.image_binning

    def set_stagematrix(self, stagematrix: 'np.array[2,2]'):
        """Set the unbinned stage matrix calibration matrix."""
        self.stagematrix = stagematrix * self.image_binning

    def set_stagematrix_from_serialem_calib(self, StageToCameraMatrix: str):
        """Set the stagematrix directly from the SerialEM calibration.

        Copy the line from `SerialEMcalibrations.txt`

        Looks something like:
        `StageToCameraMatrix 10 0 8.797544 0.052175 0.239726 8.460119   0.741238   100`

        Where the last number is the magnification. Also estimates the `pixelsize`.
        """
        inp = StageToCameraMatrix.split()[3:7]
        values = [float(val) for val in inp]
        stagematrix_inv = np.array(values).reshape(2, 2)
        stagematrix = np.linalg.inv(stagematrix_inv) * 1000

        self.set_stagematrix(stagematrix)

        pixelsize = np.abs(stagematrix.trace() / 2)  # estimate pixelsize from stagematrix
        self.set_pixelsize(pixelsize)

        return self.stagematrix

    def get_difference_vector(self,
                              idx0: int,
                              idx1: int,
                              shift: list,
                              overlap_k: float = 1.0,
                              verbose: bool = False,
                              ) -> list:
        """Calculate the pixel distance between 2 images using the calculate
        pixel shift from cross correlation.

        Parameters
        ----------
        idx0, idx1 : int
            Grid coordinate of im0 and im0, defining their relative position
        shift : list
            The offset between the 2 strips of the images used for cross correlation
        overlap_k : float
            Extend the overlap by this factor, may help with the cross correlation
            For example, if the overlap is 50 pixels, `overlap_k=1.5` will extend the
            strips used for cross correlation to 75 pixels.

        Returns
        -------
        difference_vector : np.array[1,2]
            Vector describing the pixel offset between the 2 images
        """
        res_x, res_y = self.image_shape
        overlap_x = int(self.overlap_x * overlap_k)
        overlap_y = int(self.overlap_y * overlap_k)

        vect = np.array(idx1) - np.array(idx0)
        vect = vect * np.array((res_x - overlap_x, res_y - overlap_y))

        difference_vector = vect + shift

        if verbose:
            print(f'Vector from indices: {vect}')
            print(f'Shift: {shift}')
            print(f'Difference vector: {difference_vector}')

        return difference_vector

    def calculate_difference_vectors(self,
                                     threshold: float = 'auto',
                                     overlap_k: float = 1.0,
                                     max_shift: int = 200,
                                     method: str = 'skimage',
                                     segment: bool = False,
                                     plot: bool = False,
                                     verbose: bool = False,
                                     ) -> dict:
        """Get the difference vectors between the neighbouring images The
        images are overlapping by some amount defined using `overlap`. These
        strips are compared with cross correlation to calculate the shift
        offset between the images.

        Parameters
        ----------
        threshold : float
            Lower for the cross correlation score to accept a shift or not
            If a shift is not accepted, the shift is set to (0, 0).
            Use the value 'auto' to automatically determine the threshold value.
            The threshold can be visualized using `.plot_fft_scores()`.
        overlap_k : float
            Extend the overlap by this factor, may help with the cross correlation
            For example, if the overlap is 50 pixels, `overlap_k=1.5` will extend the
            strips used for cross correlation to 75 pixels.
        max_shift : int
            Maximum pixel shift for difference vector to be accepted.
        segment : bool
            Segment the image using otsu's method before cross correlation. This improves
            the contrast for registration.
        method : str
            Set to skimage to use the cross correlation function from scikit-image. Seems
            to work better than the default in some cases.
        verbose : bool
            Be more verbose

        Returns
        -------
        difference_vectors : dict
            Dictionary with the pairwise shifts between the neighbouring
            images
        """
        grid = self.grid
        res_x, res_y = self.image_shape
        images = self.images

        overlap_x = int(self.overlap_x * overlap_k)
        overlap_y = int(self.overlap_y * overlap_k)

        pairs = define_pairs(grid)
        pairs = define_directions(pairs)

        self.pairs = pairs

        slices = make_slices(overlap_x, overlap_y)

        self.slices = slices

        results = {}

        for i, pair in enumerate(tqdm(pairs)):
            seq0 = pair['seq0']
            seq1 = pair['seq1']

            side0 = pair['side0']
            side1 = pair['side1']
            idx0 = pair['idx0']
            idx1 = pair['idx1']
            im0 = images[seq0]
            im1 = images[seq1]

            if plot and False:
                plot_images(im0, im1, seq0, seq1, side0, side1, idx0, idx1)

            # If the pair of images has already been compared, copy that result instead
            if (seq1, seq0) in results:
                if verbose:
                    print(f'{seq0:3d}{seq1:3d} | copy')
                result = results[seq1, seq0]
                shift = -result['shift']
                score = result['fft_score']
            else:
                if verbose:
                    print(f'{seq0:3d}{seq1:3d} | fft')
                strip0 = im0[slices[side0]]
                strip1 = im1[slices[side1]]

                if segment:
                    t0 = filters.threshold_otsu(strip0)
                    t1 = filters.threshold_otsu(strip1)
                    strip0 = strip0 > t0
                    strip1 = strip1 > t1
                    # print(f"Thresholds: {t1} {t1}")

                if method == 'skimage':  # method = skimage.registration.phase_cross_correlation
                    shift, error, phasediff = phase_cross_correlation(strip0, strip1, return_error=True)
                    fft = np.ones_like(strip0)
                    score = 1 - error**0.5
                else:
                    shift, fft = translation(strip0, strip1, return_fft=True)
                    score = fft.max()

                if plot:
                    plot_fft(strip0, strip1, shift, fft, side0, side1)

            shift = np.array(shift)

            results[seq0, seq1] = {
                'shift': shift,
                'idx0': idx0,
                'idx1': idx1,
                'overlap_k': overlap_k,
                'fft_score': np.nan_to_num(score, nan=0.0),
            }

            # if plot:
            #     plot_shifted(im0, im1, difference_vector, seq0, seq1, idx0, idx1, res_x, res_y)

        self.raw_difference_vectors = results

        difference_vectors = self.filter_difference_vectors(threshold=threshold,
                                                            verbose=verbose,
                                                            max_shift=max_shift)

        self.difference_vectors = difference_vectors
        self.weights = {k: v['fft_score'] for k, v in results.items()}

        return difference_vectors

    def filter_difference_vectors(self,
                                  threshold: float = 'auto',
                                  max_shift: int = 200,
                                  verbose: bool = True,
                                  plot: bool = True,
                                  ) -> dict:
        """Filter the raw difference vectors based on their fft scores.

        Parameters
        ----------
        threshold : float
            Lower for the cross correlation score to accept a shift or not
            If a shift is not accepted, the shift is set to (0, 0).
            Use the value 'auto' to automatically determine the threshold value.
            The threshold can be visualized using `.plot_fft_scores()`.
        max_shift : int
            Maximum pixel shift for difference vector to be accepted.
        verbose : bool
            Be more verbose
        plot : bool
            Plot the difference vectors
        """
        results = self.raw_difference_vectors

        if threshold == 'auto':
            scores = [item['fft_score'] for item in results.values()]
            threshold = find_threshold(scores)

        self.fft_threshold = threshold

        out = {}
        for i, (key, item) in enumerate(results.items()):
            score = item['fft_score']
            seq0, seq1 = key
            idx0 = item['idx0']
            idx1 = item['idx1']
            overlap_k = item['overlap_k']
            shift = item['shift']
            include = False

            if score < threshold:
                new_shift = np.array((0, 0))
                msg = '-> Below threshold!'
            elif np.linalg.norm(shift) > max_shift:
                new_shift = np.array(0.0)
                msg = '-> Too large!'
            else:
                new_shift = item['shift']
                msg = '-> :-)'
                include = True

            if verbose:
                t0, t1 = shift
                print(f'Pair {seq0:2d}:{idx0} - {seq1:2d}:{idx1} -> S: {score:.4f} -> Shift: {t0:4} {t1:4} {msg}')

            if include:
                out[seq0, seq1] = self.get_difference_vector(idx0,
                                                             idx1,
                                                             new_shift,
                                                             overlap_k=overlap_k,
                                                             verbose=False)

        return out

    def plot_shifts(self) -> None:
        """Plot the pixel shifts from the cross correlation."""
        shifts = np.array([item['shift'] for item in self.raw_difference_vectors.values()])
        scores = np.array([item['fft_score'] for item in self.raw_difference_vectors.values()])
        t0, t1 = shifts.T
        plt.scatter(t0, t1, c=scores, marker='+')
        plt.xlabel('Shift X (px)')
        plt.ylabel('Shift Y (px)')
        plt.title('Pixel shifts (color = FFT score)')
        plt.axis('equal')
        plt.colorbar()
        plt.show()

    def plot_fft_scores(self) -> None:
        """Plot the distribution of fft scores for the cross correlation."""
        scores = [item['fft_score'] for item in self.raw_difference_vectors.values()]
        shifts = np.array([item['shift'] for item in self.raw_difference_vectors.values()])
        amplitudes = np.linalg.norm(shifts, axis=1)
        auto_thresh = find_threshold(scores)
        used_thresh = self.fft_threshold
        plt.axhline(auto_thresh, lw=0.5, color='red', label=f'Suggested threshold={auto_thresh:.4f}')
        plt.axhline(used_thresh, lw=0.5, color='green', label=f'Actual threshold={used_thresh:.4f}')
        plt.scatter(np.arange(len(scores)), sorted(scores), c=amplitudes, marker='.')
        plt.colorbar()
        plt.title('FFT scores (color = pixel shift)')
        plt.xlabel('Index')
        plt.ylabel('Score')
        plt.legend()
        plt.show()

    def calculate_montage_coords(self) -> list:
        """Get the coordinates for each section based on the gridspec only (not
        optimized)

        Returns
        -------
        coords : np.array[-1, 2]
            Coordinates for each section in the montage map
        """

        res_x, res_y = self.image_shape
        grid = self.grid

        overlap_x = self.overlap_x
        overlap_y = self.overlap_y

        # make starting values
        vect = np.array((res_x - overlap_x, res_y - overlap_y))
        vects = []

        for i, idx in enumerate(sorted_grid_indices(grid)):
            x0, y0 = vect * idx
            vects.append((x0, y0))

        vects = np.array(vects)

        self.coords = vects

        return vects

    def optimize_montage_coords(self,
                                method: str = 'leastsq',
                                skip: tuple = (),
                                verbose: bool = False,
                                plot: bool = False,
                                ) -> list:
        """Use the difference vectors between each pair of images to calculate
        the optimal coordinates for each section using least-squares
        minimization.

        Parameters
        ----------
        method : str
            Least-squares minimization method to use (lmfit)
        skip : tuple
            List of integers of frames to ignore
        verbose : bool
            Be more verbose
        plot : bool
            Plot the original and optimized pixel coordinates

        Returns
        -------
        coords : np.array[-1, 2]
            Optimized coordinates for each section in the montage map
        """
        if not hasattr(self, 'coords'):
            self.calculate_montage_coords()
        vects = self.coords

        difference_vectors = self.difference_vectors
        weights = self.weights

        res_x, res_y = self.image_shape
        grid = self.grid
        grid_x, grid_y = grid.shape
        n_gridpoints = grid_x * grid_y

        # determine which frames items have neighours
        has_neighbours = {i for key in difference_vectors.keys() for i in key}

        # setup parameters
        params = lmfit.Parameters()

        middle_i = int(n_gridpoints / 2)  # Find index of middlemost item
        for i, row in enumerate(vects):
            if i in skip:
                vary = False
            elif i not in has_neighbours:
                vary = False
            else:
                vary = (i != middle_i)  # anchor on middle frame
            params.add(f'C{i}{0}', value=row[0], vary=vary, min=row[0] - res_x / 2, max=row[0] + res_x / 2)
            params.add(f'C{i}{1}', value=row[1], vary=vary, min=row[1] - res_y / 2, max=row[1] + res_y / 2)

        def obj_func(params, diff_vects, weights):
            V = np.array([v.value for k, v in params.items() if k.startswith('C')]).reshape(-1, 2)

            # Minimization function from 2.2
            out = []
            for i, j in diff_vects.keys():
                if i in skip:
                    continue
                if j in skip:
                    continue

                diffij = diff_vects[i, j]
                x = V[j] - V[i] - diffij
                weight = weights[i, j]
                out.append(x * weight)

            return np.array(out)

        args = (difference_vectors, weights)
        res = lmfit.minimize(obj_func, params, args=args, method=method)

        lmfit.report_fit(res, show_correl=verbose, min_correl=0.8)

        params = res.params
        Vn = np.array([v.value for k, v in params.items() if k.startswith('C')]).reshape(-1, 2)

        offset = min(Vn[:, 0]), min(Vn[:, 1])
        coords = Vn - offset

        self.optimized_coords = coords

        if plot:
            # center on average coordinate to better show displacement
            c1 = self.coords - np.mean(self.coords, axis=0)
            c2 = self.optimized_coords - np.mean(self.optimized_coords, axis=0)
            plt.title(f'Shifts from minimization (`{method}`)\nCentered on average position')
            plt.scatter(*c1.T, label='Original', marker='+')
            plt.scatter(*c2.T, label='Optimized', marker='+')
            # for (x1,y1), (x2, y2) in zip(m.coords, m.optimized_coords):
            #     arrow = plt.arrow(x1, x2, x2-x1, y2-y1)
            plt.axis('equal')
            plt.legend()

        return coords

    def _montage_patches(self, coords, binning=1):
        montage_patches = []
        for i, coord in enumerate(coords):
            image = self.images[i]
            patch = MontagePatch(image, coord, binning=binning)
            montage_patches.append(patch)
        return montage_patches

    def stitch(self,
               method: str = None,
               binning: int = 1,
               optimized: bool = True):
        """Stitch the images together using the given list of pixel coordinates
        for each section.

        Parameters
        ----------
        method : str
            Choices: [None, 'weighted', 'average']
            With `weighted`, the intensity contribution is weighted by
            the distance from the center of the image. With 'average',
            the images are averaged, and 'None' simply places the patches
            in sequential order, overwriting previous data.
        binning : int
            Bin the Montage image by this factor
        optimized : bool
            Use optimized coordinates if they are available [default = True]

        Return
        ------
        stitched : np.array
            Stitched image
        """
        if optimized:
            try:
                coords = self.optimized_coords
            except AttributeError:
                coords = self.coords
        else:
            coords = self.coords

        grid = self.grid
        images = self.images
        nx, ny = grid.shape
        res_x, res_y = self.image_shape
        n_images = len(images)

        assert nx * ny == n_images, f'Number of images ({n_images}) does not match number of grid shape ({nx}x{ny})!'

        c = coords.astype(int)
        stitched_x, stitched_y = c.max(axis=0) - c.min(axis=0)
        stitched_x += res_x
        stitched_y += res_y

        stitched = np.zeros((int(stitched_x / binning),
                             int(stitched_y / binning)),
                            dtype=np.float32)

        if method in ('average', 'weighted'):
            n_images = np.zeros_like(stitched)

            if method == 'weighted':
                weight = weight_map((int(res_x / binning),
                                     int(res_y / binning)),
                                    method='circle')

        montage_patches = self._montage_patches(coords, binning=binning)
        for i, patch in enumerate(montage_patches):
            im = patch.image
            x0 = patch.x0
            y0 = patch.y0
            x1 = patch.x1
            y1 = patch.y1

            if method == 'average':
                stitched[x0:x1, y0:y1] += im
                n_images[x0:x1, y0:y1] += 1
            elif method == 'weighted':
                stitched[x0:x1, y0:y1] += im * weight
                n_images[x0:x1, y0:y1] += weight
            else:
                stitched[x0:x1, y0:y1] = im

        if method in ('average', 'weighted'):
            n_images = np.where(n_images == 0, 1, n_images)
            stitched /= n_images

        self.stitched = stitched
        self.centers = coords + np.array((res_x, res_y)) / 2
        self.stitched_binning = binning
        self.montage_patches = montage_patches

        return stitched

    def plot(self, ax=None, vmax: int = None, labels: bool = True):
        """Plots the stitched image.

        Parameters
        ----------
        ax : matplotlib.Axis
            Matplotlib axis to plot on.
        """
        stitched = self.stitched

        if not ax:
            fig, ax = plt.subplots(figsize=(10, 10))

        grid = self.grid
        indices = sorted_grid_indices(grid)
        montage_patches = self.montage_patches
        for i, patch in enumerate(montage_patches):
            idx = indices[i]
            txt = f'{i}\n{idx}'

            if labels:
                # NOTE that y/x are flipped for display in matplotlib ONLY
                ax.text((patch.y0 + patch.y1) / 2,
                        (patch.x0 + patch.x1) / 2,
                        txt,
                        color='red',
                        fontsize=18,
                        ha='center',
                        va='center',
                        )
            rect = patches.Rectangle([patch.y0, patch.x0],
                                     patch.res_x,
                                     patch.res_y,
                                     linewidth=0.5,
                                     edgecolor='r',
                                     facecolor='none',
                                     )
            ax.add_patch(rect)

        ax.imshow(stitched, vmax=vmax)
        ax.set_title('Stitched image')

        if not ax:
            plt.show()
        else:
            return ax

    def export(self, outfile: str = 'stitched.tiff') -> None:
        """Export the stitched image to a tiff file.

        Parameters
        ----------
        outfile : str
            Name of the image file.
        """
        raise NotImplementedError

    def to_nav(self,
               fn: str = 'stitched.nav',
               coords: list = None,
               kind: str = 'pixel'):
        """Write montage to a SerialEM .nav file.
        NOTE: The stage coordinates in the .nav file are not reliable.

        Parameters
        ----------
        coords : np.array
            List of pixel / stagecoords
        kind : str
            Specify whether the coordinates are pixel or stage coordinates,
            must be one of `pixel` or `stage`.

        Returns
        -------
        map_item : `MapItem`
            Map item corresponding to the stitched image. Any coords
            specified are accessed as a dict of markers
            through `map_item.markers`.
        """
        from pyserialem import MapItem, write_nav_file

        stem = fn.rsplit('.', 1)[-1]
        fn_mrc = stem + '.mrc'
        f = mrcfile.new(fn_mrc, data=self.stitched, overwrite=True)
        f.close()

        map_scale_mat = np.linalg.inv(self.stagematrix)

        d = {
            'StageXYZ': [0, 0, 0],
            'MapFile': fn_mrc,
            'MapSection': 0,
            'MapBinning': self.image_binning,
            'MapMagInd': self.abs_mag_index,
            'MapScaleMat': map_scale_mat.flatten().tolist(),
            'MapWidthHeight': self.stitched.shape,
        }

        map_item = MapItem.from_dict(d)

        if coords is not None:
            markers = map_item.add_marker_group(coords, kind=kind)

        write_nav_file(fn, map_item, *map_item.markers.values())

        return map_item

    def pixel_to_stagecoord(self,
                            px_coord: tuple,
                            stagematrix=None,
                            plot=False,
                            ) -> 'np.array':
        """Takes a pixel coordinate and transforms it into a stage coordinate.

        Parameters
        ----------
        stage_coords : np.array (nx2)
            List of stage coordinates in nm.
        stagematrix : np.array (2x2)
            Stage matrix to convert from pixel to stage coordinates
        plot : bool
            Visualize the pixelcoordinates on the stitched images

        Returns
        -------
        np.array (nx2)
            Stage coordinates (nm) corresponding to pixel coordinates
            given in the stitched image.
        """
        if stagematrix is None:
            stagematrix = self.stagematrix

        stagematrix = self.stagematrix

        px_coord = np.array(px_coord) * self.stitched_binning
        cx, cy = np.dot(np.array(self.image_shape) / 2, stagematrix)

        diffs = np.linalg.norm((self.centers - px_coord), axis=1)
        j = np.argmin(diffs)

        image_pixel_coord = self.coords[j]
        image_stage_coord = self.stagecoords[j]

        tx, ty = np.dot(px_coord - image_pixel_coord, stagematrix)

        stage_coord = np.array((tx, ty)) + image_stage_coord - np.array((cx, cy))
        stage_coord = stage_coord.astype(int)  # round to integer

        if plot:
            img = self.images[j]
            plt.imshow(img)
            plot_x, plot_y = px_coord - image_pixel_coord
            plt.scatter(plot_y, plot_x, color='red')
            plt.text(plot_y, plot_x, ' P', fontdict={'color': 'red', 'size': 20})
            plt.title(f'Image coord: {image_stage_coord}\nP: {stage_coord}\nLinked to: {j}')
            plt.show()

        return stage_coord

    def pixel_to_stagecoords(self, pixelcoords, stagematrix=None, plot: bool = False):
        """Convert a list of pixel coordinates into stage coordinates. Uses
        `.pixel_to_stagecoord`

        Parameters
        ----------
        pixel_coords : np.array (nx2)
            List of pixel coordinates.
        stagematrix : np.array (2x2)
            Stage matrix to convert from pixel to stage coordinates
        plot : bool
            Visualize the pixelcoordinates on the stitched images

        Returns
        -------
        stage_coords : np.array (nx2)
            stage coordinates corresponding to the given pixel coordinates on the stitched image.
        """
        if stagematrix is None:
            stagematrix = self.stagematrix

        f = self.pixel_to_stagecoord
        stage_coords = np.array([f(px_coord, stagematrix=stagematrix) for px_coord in pixelcoords])

        if plot:
            plot_x, plot_y = stage_coords.T
            plt.scatter(plot_x, plot_y, color='red', marker='.')
            plt.title(f'Stage coordinates')
            plt.xlabel('X (nm)')
            plt.xlabel('Y (nm)')
            plt.show()

        return stage_coords

    def stage_to_pixelcoord(self,
                            stage_coord: tuple,
                            stagematrix=None,
                            plot: bool = False,
                            ) -> 'np.array':
        """Takes a stage coordinate and transforms it into a pixel coordinate.

        Note that this is not the inverse of `.pixel_to_stagecoord`, as this function
        finds the closest image to 'hook onto', and calculates the pixel coordinate
        with that image as the reference position.

        Parameters
        ----------
        stage_coords : np.array (nx2)
            List of stage coordinates in nm.
        stagematrix : np.array (2x2)
            Stage matrix to convert from pixel to stage coordinates
        plot : bool
            Visualize the pixelcoordinates on the stitched images

        Returns
        -------
        px_cord : np.array (1x2)
            Pixel coordinates corresponding to the stitched image.
        """
        if stagematrix is None:
            stagematrix = self.stagematrix

        mat = stagematrix
        mati = np.linalg.inv(stagematrix)

        center_offset = np.dot(np.array(self.image_shape) / 2, mat)
        stage_coord = np.array(stage_coord)

        diffs = np.linalg.norm((self.stagecoords - stage_coord), axis=1)
        j = np.argmin(diffs)

        image_stage_coord = self.stagecoords[j]
        image_pixel_coord = self.coords[j].copy()

        # The image pixel coordinate `image_pixel_coord` corresponds to the corner,
        # but the stage coord `image_stage_coord` at the center of the image.
        # `px_coord` is the relative offset added to the corner pixel coordiante of the image
        px_coord = np.dot(stage_coord - image_stage_coord + center_offset, mati) + image_pixel_coord

        px_coord /= self.stitched_binning
        px_coord = px_coord.astype(int)

        if plot:
            plot_x, plot_y = px_coord
            plt.imshow(self.stitched)
            plt.scatter(plot_y, plot_x, color='red', marker='.')
            plt.text(plot_y, plot_x, ' P', fontdict={'color': 'red', 'size': 20})
            plt.title(f'P: {px_coord}\nStage: {stage_coord}\nLinked to: {j}')
            plt.show()

        return px_coord

    def stage_to_pixelcoords(self,
                             stage_coords: tuple,
                             stagematrix=None,
                             plot: bool = False,
                             ) -> 'np.array':
        """Convert a list of stage coordinates into pixelcoordinates. Uses
        `.stage_to_pixelcoord`

        Note that this is not the inverse of `.pixel_to_stagecoord`, as this function
        finds the closest image to 'hook onto', and calculates the pixel coordinate
        with that image as the reference position.

        Parameters
        ----------
        stage_coords : np.array (nx2)
            List of stage coordinates in nm.
        stagematrix : np.array (2x2)
            Stage matrix to convert from pixel to stage coordinates
        plot : bool
            Visualize the pixelcoordinates on the stitched images

        Returns
        -------
        px_coords : np.array (nx2)
            Pixel coordinates corresponding to the stitched image.
        """
        f = self.stage_to_pixelcoord
        px_coords = np.array([f(stage_coord, stagematrix=stagematrix) for stage_coord in stage_coords])

        if plot:
            plot_x, plot_y = px_coords.T
            plt.imshow(self.stitched)
            plt.scatter(plot_y, plot_x, color='red', marker='.')
            plt.title(f'Pixel coordinates mapped on stitched image')
            plt.show()

        return px_coords

    def find_holes(self,
                   diameter: float = None,
                   tolerance: float = 0.1,
                   pixelsize: float = None,
                   threshold: int = None,
                   plot: bool = False,
                   ) -> tuple:
        """Find grid holes in the montage image.

        Parameters
        ----------
        diameter : float
            In nm, approximate diameter of squares/grid holes. If it
            it is not specified, take the median diameter of all props.
        tolerance : float
            Tolerance in % how far the calculate diameter can be off
        pixelsize : float
            Unbinned tile pixelsize in nanometers
        threshold: int
            Threshold to use for segmentation. By default the `otsu` method
            is used to find a suitable threshold.
        plot : bool
            Plot the segmentation results and coordinates using Matplotlib

        Returns
        -------
        stagecoords : np.array, imagecoords : np.array
            Return both he stage and imagecoords as numpy arrays
        """
        from skimage import filters
        from skimage import morphology
        from skimage.measure import regionprops

        stitched = self.stitched

        if not threshold:
            threshold = filters.threshold_otsu(stitched)
        selem = morphology.disk(10)
        seg = morphology.binary_closing(stitched > threshold, selem=selem)

        labeled, _ = ndimage.label(seg)
        props = regionprops(labeled)

        stitched_binning = self.stitched_binning

        if not pixelsize:
            pixelsize = self.pixelsize * stitched_binning
        else:
            pixelsize *= stitched_binning

        if diameter is None:
            diameter = np.median([(prop.area ** 0.5) * pixelsize for prop in props])
            print(f'Diameter: {diameter:.0f} nm')

        max_val = tolerance * diameter

        stagecoords = []
        imagecoords = []

        allds = []  # all diameters
        selds = []  # selected diameters

        for prop in props:
            x, y = prop.centroid

            d = (prop.area ** 0.5) * pixelsize
            allds.append(d)

            if abs(d - diameter) < max_val:
                stagecoord = self.pixel_to_stagecoord((x, y))
                stagecoords.append(stagecoord)
                imagecoords.append((x, y))
                selds.append(d)

        stagecoords = np.array(stagecoords)
        imagecoords = np.array(imagecoords)

        if plot:
            fig, (ax0, ax1, ax2) = plt.subplots(ncols=3, figsize=(12, 4))

            ax0.set_title(f'Segmentation (z>{threshold:.0f})')
            ax0.imshow(seg)

            ax1.set_title('Image coords')
            ax1.imshow(stitched)

            ax2.set_title('Stage coords')

            try:
                plot_x, plot_y = np.array(imagecoords).T
                ax1.scatter(plot_y, plot_x, marker='+', color='r')
            except ValueError:
                pass

            try:
                plot_x, plot_y = np.array(stagecoords).T
                ax2.scatter(plot_x / 1000, plot_y / 1000, marker='+')
                ax2.scatter(0, 0, marker='+', color='red', label='Origin')
                ax2.legend()
                ax2.axis('equal')
                ax2.set_xlabel('X (μm)')
                ax2.set_ylabel('Y (μm)')
            except ValueError:
                pass

        prc = np.percentile
        mdn = np.median
        print(f'All hole diameters     50%: {mdn(allds):6.0f} | 5%: {prc(allds, 5):6.0f} | 95%: {prc(allds, 95):6.0f}')
        if len(selds) > 0:
            print(f'Selected hole diameter 50%: {mdn(selds):6.0f} | 5%: {prc(selds, 5):6.0f} | 95%: {prc(selds, 95):6.0f}')
        else:
            print(f'Selected hole diameter 50%: {"-":>6s} | 5%: {"-":>6s} | 95%: {"-":>6s}')

        self.feature_coords_stage = stagecoords
        self.feature_coords_image = imagecoords

        return stagecoords, imagecoords


if __name__ == '__main__':
    pass



from .montage import Montage
from .reader import MapItem
from .reader import NavItem
from .reader import read_mdoc_file
from .reader import read_nav_file
from .reader import stitch_map_items
from .reader import write_nav_file

__version__ = '0.3.2'



import numpy as np
import pytest

import pyserialem as pysem


@pytest.fixture
def map_item(drc):
    fn = drc.parent / 'data' / 'nav.nav'
    items = pysem.read_nav_file(fn)
    map_item = items[0]
    assert map_item.kind == 'Map'
    return map_item


def test_coordinate_convert(map_item):
    assert map_item.stagematrix.shape == (2, 2)

    xy = (1, 1)

    new_xy = map_item.pixel_to_stagecoords(xy)

    assert len(new_xy) == 2

    old_xy = map_item.stage_to_pixelcoords(new_xy)

    assert len(new_xy) == 2
    assert np.allclose(xy, old_xy)


@pytest.mark.skip(reason='No access to data')
def test_load_image(map_item, drc):
    img = map_item.load_image(drc.parent / 'data')

    assert isinstance(img, np.ndarray)
    assert img.shape == tuple(map_item.MapWidthHeight)


def test_add_markers(map_item):
    n = 10

    coords = np.arange(n * 2).reshape(n, 2)
    nav_items = map_item.add_marker_group(coords)

    assert len(nav_items) == n

    pixelcoords = map_item.markers_as_pixel_coordinates()
    assert pixelcoords.shape == (n, 2)
    assert np.allclose(pixelcoords, coords)

    stagecoords = map_item.markers_as_stage_coordinates()
    assert stagecoords.shape == (n, 2)

    nav_item = nav_items[0]
    assert nav_item.kind == 'Marker'

    map_item.update_markers(nav_item)

    assert len(map_item.markers) == n

    map_item.set_markers(nav_item)

    assert len(map_item.markers) == 1


def test_to_from_dict(map_item):
    d = map_item.to_dict()

    new_map_item = pysem.MapItem.from_dict(d)

    e = new_map_item.to_dict()

    assert isinstance(e, dict)

    # cycle twice to get rid of rounding errors
    new_map_item = pysem.MapItem.from_dict(e)

    f = new_map_item.to_dict()

    assert f == e


def test_validate(map_item):
    map_item.validate()

    del map_item.StageXYZ

    with pytest.raises(KeyError):
        map_item.validate()


def test_tag_loading(map_item, drc):
    n = 10

    coords = np.arange(n * 2).reshape(n, 2)
    nav_items = map_item.add_marker_group(coords, acquire=True)

    out = drc.parent / 'data' / 'out2.nav'
    pysem.write_nav_file(out, map_item, *nav_items)

    items = pysem.read_nav_file(out, acquire_only=False)
    assert len(items) == 1 + n

    items = pysem.read_nav_file(out, acquire_only=True)
    assert len(items) == n


def test_nav_item(map_item):
    coords = np.arange(2).reshape(1, 2)
    nav_items = map_item.add_marker_group(coords)
    nav_item = nav_items[0]

    assert isinstance(nav_item.stage_x, float)
    assert isinstance(nav_item.stage_y, float)
    assert isinstance(nav_item.stage_z, float)
    assert len(nav_item.color_rgba) == 4

    assert nav_item.color_str == 'red'

    d = nav_item.to_dict()

    new_nav_item = pysem.NavItem(d, tag='test')
    assert new_nav_item.tag == 'test'



from pathlib import Path

import numpy as np
import tifffile
import yaml
from numpy.fft import fft2
from numpy.fft import ifft2


def bin_ndarray(ndarray, new_shape=None, binning=1, operation='mean'):
    """Bins an ndarray in all axes based on the target shape, by summing or
    averaging. If no target shape is given, calculate the target shape by the
    given binning.

    Number of output dimensions must match number of input dimensions and
        new axes must divide old ones.

    Example
    -------
    >>> m = np.arange(0,100,1).reshape((10,10))
    >>> n = bin_ndarray(m, new_shape=(5,5), operation='sum')
    >>> print(n)

    [[ 22  30  38  46  54]
     [102 110 118 126 134]
     [182 190 198 206 214]
     [262 270 278 286 294]
     [342 350 358 366 374]]
    """
    if not new_shape:
        shape_x, shape_y = ndarray.shape
        new_shape = int(shape_x / binning), int(shape_y / binning)

    if new_shape == ndarray.shape:
        return ndarray

    operation = operation.lower()
    if operation not in ['sum', 'mean']:
        raise ValueError('Operation not supported.')
    if ndarray.ndim != len(new_shape):
        raise ValueError(f'Shape mismatch: {ndarray.shape} -> {new_shape}')
    compression_pairs = [(d, c // d) for d, c in zip(new_shape,
                                                     ndarray.shape)]
    flattened = [l for p in compression_pairs for l in p]
    ndarray = ndarray.reshape(flattened)
    for i in range(len(new_shape)):
        op = getattr(ndarray, operation)
        ndarray = op(-1 * (i + 1))
    return ndarray


def translation(im0,
                im1,
                limit_shift: bool = False,
                return_fft: bool = False,
                ):
    """Return translation vector to register images.

    Parameters
    ----------
    im0, im1 : np.array
        The two images to compare
    limit_shift : bool
        Limit the maximum shift to the minimum array length or width.
    return_fft : bool
        Whether to additionally return the cross correlation array between the 2 images

    Returns
    -------
    shift: list
        Return the 2 coordinates defining the determined image shift
    """
    f0 = fft2(im0)
    f1 = fft2(im1)
    ir = abs(ifft2((f0 * f1.conjugate()) / (abs(f0) * abs(f1))))
    shape = ir.shape

    if limit_shift:
        min_shape = min(shape)
        shift = int(min_shape / 2)
        ir2 = np.roll(ir, (shift, shift), (0, 1))
        ir2 = ir2[:min_shape, :min_shape]
        t0, t1 = np.unravel_index(np.argmax(ir2), ir2.shape)
        t0 -= shift
        t1 -= shift
    else:
        t0, t1 = np.unravel_index(np.argmax(ir), shape)
        if t0 > shape[0] // 2:
            t0 -= shape[0]
        if t1 > shape[1] // 2:
            t1 -= shape[1]

    if return_fft:
        return [t0, t1], ir
    else:
        return [t0, t1]



import numpy as np
import pytest

import pyserialem as pysem


def test_reader_writer(drc):
    fn = drc.parent / 'data' / 'nav.nav'

    items = pysem.read_nav_file(fn)

    assert isinstance(items, list)
    assert len(items) == 1

    out = drc.parent / 'data' / 'out.nav'

    pysem.write_nav_file(out, *items)

    assert out.exists()

    pysem.read_nav_file(out)

    assert isinstance(items, list)
    assert len(items) == 1


def test_mdoc_reader(drc):
    fn = drc.parent / 'data' / 'gm.mrc.mdoc'

    items = pysem.read_mdoc_file(fn)

    assert len(items) == 26
    assert isinstance(items[0], dict)

    kind = 'ZValue'
    items = pysem.read_mdoc_file(fn, only_kind=kind)
    assert len(items) == 25
    assert items[0]['kind'] == kind

    kind = 'MontSection'
    items = pysem.read_mdoc_file(fn, only_kind=kind)
    assert len(items) == 1
    assert items[0]['kind'] == kind



import random
import re
from collections import defaultdict
from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import patches

from .utils import bin_ndarray


__all__ = [
    'bin_ndarray',
    'stitch_map_items',
    'NavItem',
    'MapItem',
    'block2dict',
    'block2nav',
    'read_nav_file',
    'write_nav_file',
    'read_mdoc_file',
]

# int
INTEGER = ('Color', 'NumPts', 'Draw', 'Regis',
           'MapMontage', 'MapSection', 'MapBinning', 'MapMagInd',
           'MapCamera', 'ShutterMode', 'MapSpotSize',
           'MapSlitIn', 'ImageType', 'MontUseStage',
           'MapProbeMode', 'MapLDConSet', 'Type', 'GroupID',
           'MapID', 'PieceOn', 'Acquire', 'DrawnID',
           'MontBinning', 'SamePosId', 'OrigReg',
           # mdoc
           'SpotSize',
           'Binning', 'CameraIndex', 'DividedBy2', 'MagIndex',
           'Magnification', 'ProbeMode', 'MoveStage',
           'Alpha', 'ImageSize', 'DataMode', 'Montage',
           'ImageSeries', 'UsingCDS', 'LowDoseConSet', 'NumSubFrames',
           # other
           'Corner', 'Imported', 'K2ReadMode', 'MapAlpha',
           'PolyID', 'RealignReg', 'RealignedID', 'RegPt',
           'RegisteredToID', 'RotOnLoad',
           # DE-12
           'DE12-TotalNumberOfFrames',
           'DE12-FramesPerSecond',
           'DE12-CameraPosition',
           'DE12-ProtectionCoverMode',
           'DE12-ProtectionCoverOpenDelay(ms)',
           'DE12-TemperatureDetector(C)',
           'DE12-SensorModuleSerialNumber',
           'DE12-SensorReadoutDelay(ms)',
           'DE12-IgnoredFramesInSummedImage',
           )

# float
FLOAT = ('MapExposure', 'MapIntensity', 'MapTiltAngle', 'MapSettling',
         # .mdoc
         'StageZ', 'PixelSpacing', 'Defocus', 'RotationAngle',
         'CountsPerElectron', 'TargetDefocus', 'TiltAngle', 'ExposureTime',
         'DriftSettling', 'Intensity', 'ExposureDose', 'PriorRecordDose',
         # other
         'DefocusOffset', 'FocusAxisPos', 'MapSlitWidth',
         # DE-12
         'DE12-ServerSoftwareVersion',
         'DE12-PreexposureTime(s)',
         'DE12-FaradayPlatePeakReading(pA/cm2)',
         )

# str
STRING = ('MapFile', 'Note',
          # .mdoc
          'DateTime', 'ImageFile', 'NavigatorLabel',
          'SubFramePath', 'ChannelName',
          )

# list, float
FLOAT_LIST = ('StageXYZ', 'RawStageXY', 'MapScaleMat', 'XYinPc',
              'PtsX', 'PtsY', 'StageXYZ', 'MapMinMaxScale',
              # .mdoc
              'StagePosition', 'MinMaxMean', 'XedgeDxyVS', 'YedgeDxyVS',
              'XedgeDxy', 'YedgeDxy', 'ImageShift', 'BufISXY',
              # other
              'BklshXY', 'FocusOffsets', 'LocalErrXY', 'NetViewShiftXY',
              'RealignErrXY', 'ViewBeamShiftXY', 'ViewBeamTiltXY',
              'SuperMontCoords', 'StageOffsets', 'FrameDosesAndNumbers',
              'FilterSlitAndLoss',
              # external
              'CoordsInMap', 'CoordsInAliMont', 'CoordsInAliMontVS', 'CoordsInPiece',
              )

# list, int
INTEGER_LIST = ('MapWidthHeight', 'MapFramesXY',
                # .mdoc
                'PieceCoordinates', 'AlignedPieceCoordsVS',
                'AlignedPieceCoords', 'MontBacklash',
                'ValidBacklash', 'CameraModes', 'FilterState',
                'ConSetUsed', 'MultishotHoleAndPosition',
                # other
                'HoleArray', 'LDAxisAngle', 'SkipHoles',
                'SuperMontXY',
                )

UNDEFINED = ()

REQUIRED_MAPITEM = ('StageXYZ', 'MapFile', 'MapSection',
                    'MapBinning', 'MapMagInd', 'MapScaleMat',
                    'MapWidthHeight', 'Color', 'Regis',
                    'Type', 'MapID', 'MapMontage', 'MapCamera',
                    'NumPts', 'PtsX', 'PtsY',
                    )


def stitch_map_items(items: list,
                     vmin: int = None,
                     vmax: int = None,
                     binning: int = 16,
                     markers: list = None,
                     scatter_kwargs: dict = {},
                     color: str = 'red',
                     label: bool = True,
                     ) -> None:
    """Take a list of MapItems and plot them at scale with respect to each
    other.

    Parameters
    ----------
    vmax : int
        Passed to plt.imshow to tune the contrast.
    vmin : int
        Passed to plt.imshow to tune the contrast.
    binning : int
        Bin the loaded images before displaying them to save
        on memory and computation times
    markers: list of arrays (Mx2)
        Must be the same length as `items`. Each item in markers
        is a numpy array (Mx2) containing x/y pixel coordinates (i.e. of
        particle positions) to plot on the stitched map.
    scatter_kwargs: dict
        Passed to `plt.scatter` to style the points passed using
        `markers`.
    color: str
        Color of all the decorations (markers, borders, text)
    label: bool
        Label each frame with the sequence number
    """
    _scatter_kwargs = {
        'marker': '.',
        'color': 'blue',
        's': 2,
    }

    _scatter_kwargs.update(scatter_kwargs)

    positions = []

    fig, ax = plt.subplots()

    try:
        tag_min = min(int(item.tag) for item in items)
    except ValueError:
        tag_min = None

    xy_coords = []

    for i, item in enumerate(items):
        sx, sy = item.stage_xy
        pos_x, pos_y = item.stage_to_pixelcoords((sx / 1000, sy / 1000))

        img = item.load_image()

        if binning:
            img = bin_ndarray(img, binning=binning)
            pos_x /= binning
            pos_y /= binning

        positions.append((pos_x, pos_y))

        shape_x, shape_y = img.shape
        shape_x = shape_x // 2
        shape_y = shape_y // 2

        im = ax.imshow(img,
                       interpolation='bilinear',
                       extent=[pos_y - shape_y, pos_y + shape_y, pos_x - shape_x, pos_x + shape_x],
                       clip_on=True,
                       vmax=vmax, vmin=vmin)

        rect = patches.Rectangle((pos_y - shape_y, pos_x - shape_x), shape_y * 2, shape_x * 2,
                                 fill=False,
                                 edgecolor=color,
                                 linewidth=1)
        ax.add_patch(rect)

        if label:
            try:
                item.tag = str(int(item.tag) - tag_min)
            except ValueError:
                tag = item.tag

            ax.text(pos_y, pos_x, tag, ha='center', va='center', color=color)

        if markers:
            xy_coord = markers[i]
            xy_coord = np.array((-1, 1)) * xy_coord / binning + np.array((pos_x, pos_y)) - np.array((-shape_x, shape_y))
            xy_coords.append(xy_coord)

    ax.scatter(0, 0, color=color, marker='+', label='Origin')
    title = f'Overview of {len(items)} frames'

    if markers:
        xy_coords = np.vstack(xy_coords)
        x, y = xy_coords.T

        n_tot = len(xy_coords)
        n_avg = len(xy_coords) / len(items)
        ax.scatter(y, x, **_scatter_kwargs, label='Particles')
        title += f'\n{n_tot} particles (average: {n_avg:.3f} / frame)'

    positions = np.array(positions)
    xmin, ymin = positions.min(axis=0)
    xmax, ymax = positions.max(axis=0)
    ax.set_ylim(xmin - shape_x, xmax + shape_x)
    ax.set_xlim(ymin - shape_x, ymax + shape_y)
    ax.axis('off')
    plt.title(title)
    plt.show()


def item_to_string(d: dict, tag: str):
    """Turn a SerialEM key/values dictionary into a .mdoc/.nav formatted
    string."""
    s = f'[Item = {tag}]\n'

    for key in sorted(d.keys()):
        val = d[key]

        try:
            if key in INTEGER:
                val = str(val)
            elif key in FLOAT:
                val = str(val)
            elif key in FLOAT_LIST:
                val = ' '.join([str(x) for x in val])
            elif key in INTEGER_LIST:
                val = ' '.join([str(x) for x in val])
        except TypeError as e:
            print(e)
            print(key, val)

        s += f'{key} = {val}\n'

    s += ''
    return s


class NavItem:
    """DataClass for SerialEM Nav items.

    Type:
        0: Marker
        1: Polygon
        2: Map
    """

    TAG_ID_ITERATOR = 1
    # MAP_ID_ITERATOR = 1

    def __init__(self, d: dict, tag: str):
        super().__init__()
        # if not "MapID" in d:
        #     d["MapID"] = NavItem.MAP_ID_ITERATOR
        #     NavItem.MAP_ID_ITERATOR += 1

        self._keys = tuple(d.keys())

        self.DrawnID = 0
        self.Acquire = 0
        self.__dict__.update(d)

        if not tag:
            tag = f'Item-{NavItem.TAG_ID_ITERATOR}'
            NavItem.TAG_ID_ITERATOR += 1

        self.tag = tag

    def __repr__(self):
        return f'{self.__class__.__name__}({self.kind}[Item = {self.tag}])'

    @property
    def kind(self) -> str:
        return ('Marker', 'Polygon', 'Map')[self.Type]

    @property
    def stage_x(self) -> float:
        return self.StageXYZ[0]

    @property
    def stage_y(self) -> float:
        return self.StageXYZ[1]

    @property
    def stage_z(self) -> list:
        return self.StageXYZ[2]

    @property
    def stage_xy(self) -> list:
        return self.StageXYZ[0:2]

    def to_string(self) -> str:
        """Convert nav item to string that can be printed to .nav file."""
        d = self.to_dict()
        return item_to_string(d, tag=self.tag)

    def to_dict(self) -> dict:
        """Convert nav item back to dictionary."""
        return {key: self.__dict__[key] for key in self._keys}

    @property
    def color_rgba(self) -> tuple:
        """Return matplotlib RGBA color."""
        return mpl.colors.to_rgba(self.color_str, alpha=None)

    @property
    def color_str(self) -> str:
        """Return color as string."""
        return ('red', 'green', 'blue', 'yellow', 'magenta', 'black')[self.Color]


class MapItem(NavItem):
    """Adds some extra methods for map items."""

    GROUP_ID_ITERATOR = random.randint(1, 90000)

    def __init__(self, *args, data_drc: str = '.', **kwargs):
        super().__init__(*args, **kwargs)

        self.validate()
        self.markers = {}
        self.data_drc = data_drc

    @property
    def stagematrix(self) -> 'np.array':
        # MapScaleMat already has binning embedded
        # MapBinning = self.MapBinning

        try:
            MontBinning = self.MontBinning
        except AttributeError:
            MontBinning = 1
        stagematrix = (1 / MontBinning) * np.array(self.MapScaleMat).reshape(2, 2)

        # FIXME: why do we need the transpose?
        stagematrix = stagematrix.T

        return stagematrix

    def pixel_to_stagecoords(self, coords: list) -> 'np.array':
        """Convert from pixel coordinates to stage coordinates."""
        coords = np.array(coords)
        cp = np.array(self.MapWidthHeight) / 2
        cs = np.array(self.StageXYZ)[0:2]
        mati = np.linalg.inv(self.stagematrix)

        return np.dot(coords - cp, mati) + cs

    def stage_to_pixelcoords(self, coords: list) -> 'np.array':
        """Convert from stage coordinates to pixel coordinates."""
        coords = np.array(coords)
        cp = np.array(self.MapWidthHeight) / 2
        cs = np.array(self.StageXYZ)[0:2]
        mat = self.stagematrix

        return np.dot(coords - cs, mat) + cp

    def load_image(self, drc: str = None) -> 'np.array':
        """Loads the image corresponding to this item."""
        import mrcfile

        if not drc:
            drc = self.data_drc
        drc = Path(drc)

        map_file = Path(self.MapFile)
        if not map_file.exists():
            map_file = drc / Path(self.MapFile).name

        m = mrcfile.mmap(map_file)
        s = self.MapSection
        if m.is_single_image() and s == 0:
            return np.array(m.data)
        else:
            return np.array(m.data[s])

    def plot_image(self, markers: bool = True, drc: str = None) -> None:
        """Plot the image including markers (optional)"""
        import matplotlib.pyplot as plt

        if markers is True:
            markers = self.markers.values()
        elif isinstance(markers, dict):
            markers = markers.values()
        elif isinstance(markers, (list, tuple, np.ndarray)):
            pass
        else:
            markers = []

        im = self.load_image()
        plt.matshow(im, vmax=np.percentile(im, 99))
        yres = self.MapWidthHeight[1]

        coords = []
        for marker in markers:
            if isinstance(marker, NavItem):
                xy = np.array([marker.stage_x, marker.stage_y])
                px, py = self.stage_to_pixelcoords(xy)
                py = yres - py
            else:
                py, px = marker

            coords.append((px, py))

        if coords:
            px, py = np.array(coords).T
            plt.plot(px, py, 'ro', markerfacecolor='none', markersize=20, markeredgewidth=2)

        plt.show()

    def add_marker(self,
                   coord: tuple,
                   kind: str = 'pixel',
                   tag: str = None,
                   acquire: bool = True,
                   ) -> 'NavItem':
        """Add pixel or stage coordinate as marker to a map item. Markers are
        linked to this `MapItem` via the `.markers` attribute.

        Parameters
        ----------
        coord : array (n x 2)
            List of X, Y pixel coordinates or stage coordinates corresponding to the navigation item.
        kind : str
            Defines the kind of coordinate supplied, must be one of `pixel` or `stage`. Stage coordinates
            are given in μm.
        tag : str
            Simple name tag for the item. It will be generated automatically if it is not given.
        acquire : bool
            Turn on the acquire flag for this item.

        Returns
        -------
        Instance of `NavItem`
        """
        if kind == 'pixel':
            py, px = coord
            yres = self.MapWidthHeight[1]
            py = yres - py
            stage_x, stage_y = self.pixel_to_stagecoords((px, py))
        else:
            stage_x, stage_y = coord

        d = {}
        try:
            d['BklshXY'] = self.BklshXY
        except AttributeError:
            d['BklshXY'] = 10, -10
        d['Color'] = 0
        d['DrawnID'] = self.MapID
        d['GroupID'] = MapItem.GROUP_ID_ITERATOR
        # d["MapID"] = 1 + self.i  # must be a unique ID under 100000 [optional, default=0]
        d['Acquire'] = int(acquire)
        d['NumPts'] = 1
        d['PtsX'] = [stage_x]
        d['PtsY'] = [stage_y]
        d['Regis'] = self.Regis
        d['StageXYZ'] = [stage_x, stage_y, self.stage_z]
        d['Type'] = 0
        if kind == 'pixel':
            d['pixel_coord'] = coord

        item = NavItem(d, tag=tag)

        self.markers[item.tag] = item

        return item

    def add_marker_group(self,
                         coords: list,
                         kind: str = 'pixel',
                         acquire: bool = True,
                         replace: bool = True,
                         ) -> list:
        """Add pixel coordinates (numpy) as markers to a map item If
        `replace==True`, replace the entire list of existing markers on this
        `MapItem` (via `.markers`).

        Parameters
        ----------
        coords : array (n x 2)
            List of X, Y pixel coordinates or stage coordinates corresponding to the navigation item.
        kind : str
            Defines the kind of coordinate supplied, must be one of `pixel` or `stage`. Stage coordinates
            are given in μm.
        acquire : bool
            Turn on the acquire flag for this item.
        replace : bool
            Replace the exisiting items on this instance of `MapItem`

        Returns
        -------
        List of `NavItem` instances
        """
        if replace:
            self.markers = {}

        ret = []
        for i, coord in enumerate(coords):
            tag = f'{self.tag}-{i}'
            item = self.add_marker(coord, kind=kind, tag=tag, acquire=True)
            ret.append(item)

        MapItem.GROUP_ID_ITERATOR += 1

        return ret

    def update_markers(self, *items):
        """Update the list of markers belonging to this `Map` with the given
        items."""
        for item in items:
            self.markers[item.tag] = item

    def set_markers(self, *items):
        """Replace the list of markers belonging to this `Map` with the given
        items."""
        self.markers = {}
        self.update_markers(*items)

    def markers_as_pixel_coordinates(self):
        """Return markers as (Mx2) array of pixel xy coordinates."""
        try:
            return np.array([marker.pixel_coord for marker in self.markers.values()])
        except AttributeError:
            raise NotImplementedError()  # This function is broken in some cases
            markers = self.markers_as_stage_coordinates()
            return self.stage_to_pixelcoords(markers)

    def markers_as_stage_coordinates(self):
        """Return markers as (Mx2) array of stage xy coordinates (µm)."""
        return np.array([marker.stage_xy for marker in self.markers.values()])

    @classmethod
    def from_dict(cls, dct, tag: str = None):
        """Construct a new map item from a dictionary.

        For the required keys, see `serialem.REQUIRED_MAPITEM

        Parameters
        ----------
        dct : dict
            Dictionary of required items
        tag : str
            Name to identify the map item by

        Returns
        -------
        map_item : MapItem
        """
        MapID = MapItem.GROUP_ID_ITERATOR
        MapItem.GROUP_ID_ITERATOR += 1

        # required items that can be generated
        map_dct = {
            'Color': 2,
            'Regis': 1,
            'Type': 2,
            'MapID': MapID,
            'MapMontage': 0,
            'MapCamera': 0,
            'NumPts': 5,       # number of points describing square around it?
            'PtsX': (-1, 1, 1, -1, -1),  # draw square around point, grid coordinates
            'PtsY': (-1, -1, 1, 1, -1),  # draw square around point, grid coordinates
        }

        map_dct.update(dct)

        if not tag:
            tag = str(MapID)

        map_item = cls(map_dct, tag=tag)
        map_item.calculate_PtsXY()

        return map_item

    def validate(self) -> None:
        """Validate the dictionary.

        Check whether all necessary keys are present
        """
        for key in REQUIRED_MAPITEM:
            if key not in self.__dict__:
                raise KeyError(f'MapItem: missing key `{key}`')

    def calculate_PtsXY(self) -> None:
        """Calculate PtsX / PtsY from the `map_item` information
        (MapWidthHeight) via `.pixel_to_stagecoords`.

        Updates the internal values.
        """
        x, y = self.MapWidthHeight

        pts = np.array((
            (0, 0),
            (x, 0),
            (x, y),
            (0, y),
            (0, 0),
        ))

        coords = self.pixel_to_stagecoords(pts)

        PtsX, PtsY = coords.T
        self.PtsX = PtsX.tolist()
        self.PtsY = PtsY.tolist()


def block2dict(block: list, kind: str = None, sequence: int = -1) -> dict:
    """Takes a text block from a SerialEM .nav file and converts it into a
    dictionary."""
    patt_split = re.compile(r'\s?=\s?')
    d = {}

    for item in block:
        key, value = re.split(patt_split, item)

        try:
            if key in INTEGER:
                value = int(value)
            elif key in FLOAT:
                value = float(value)
            elif key in STRING:
                value = str(value)
            elif key in FLOAT_LIST:
                value = [float(val) for val in value.split()]
            elif key in INTEGER_LIST:
                value = [int(val) for val in value.split()]
            elif key in UNDEFINED:
                print(item)
            else:
                print('Unknown item:', item)
        except Exception as e:
            print(e)
            print(item)
            print(key, value)
            raise

        d[key] = value

    if sequence >= 0:
        d['sequence'] = sequence
    if kind:
        d['kind'] = kind

    return d


def block2nav(block: list, tag: str = None, data_drc: str = None) -> 'NavItem':
    """Takes a text block from a SerialEM .nav file and converts it into a
    instance of `NavItem` or `MapItem`"""
    d = block2dict(block)
    kind = d['Type']

    if kind == 2:
        ret = MapItem(d, tag=tag, data_drc=data_drc)
    else:
        ret = NavItem(d, tag=tag)

    return ret


def read_nav_file(fn: str, acquire_only: bool = False) -> list:
    """Reads a SerialEM .nav file and returns a list of dictionaries containing
    nav item data.

    acquire_only: bool
        read only files with the Acquire tag set
    """

    # https://regex101.com/
    patt_match = re.compile(r'\[Item\s?=\s?([a-zA-Z0-9_-]*)\]')

    capture = False
    block = []
    items = []
    tag = ''
    drc = Path(fn).absolute().parent

    f = open(fn, 'r')
    for line in f:
        line = line.strip()
        if not line:
            continue

        m = re.match(patt_match, line)

        if m:
            if block:
                items.append(block2nav(block, tag=tag, data_drc=drc))

            # prep for next block
            tag = m.groups()[0]
            block = []
            capture = True
        elif capture:
            block.append(line)
        else:
            if line.startswith('AdocVersion'):
                pass
            elif line.startswith('LastSavedAs'):
                pass
            else:
                print(line)

    items.append(block2nav(block, tag=tag, data_drc=drc))

    if acquire_only:
        items = [item for item in items if item.Acquire]

    # associate markers with map items
    map_items = (item for item in items if item.kind == 'Map')
    markers = (item for item in items if item.kind == 'Marker')

    d = defaultdict(list)

    for marker in markers:
        d[marker.DrawnID].append(marker)

    for map_item in map_items:
        markers = d[map_item.MapID]
        map_item.update_markers(*markers)

    return items


def write_nav_file(fn: str, *items, mode='w') -> None:
    """Write list of nav items to a navigator file with filename `fn` to be
    read by SerialEM.

    `items` must be a list of NavItem / MapItem objects

    mode can be "w" to write a new file, or "a" to append
    items to an existing file
    """
    f = open(fn, mode) if fn else None
    version = '2.00'

    if mode == 'w':
        print(f'AdocVersion = {version}', file=f)
        print(f'LastSavedAs = {fn}', file=f)
        print('', file=f)

    for item in items:
        print(item.to_string(), file=f)


def read_mdoc_file(fn: str, only_kind: str = None) -> list:
    """Reads a SerialEM .mdoc file and returns a list of dictionaries
    containing supporting data.

    Parameters
    ----------
    only_kind : str
        Return only items of this kind, i.e. ZValue or MontSection (case-insensitive)

    Returns:
    --------
    List of dicts with header information from the .mdoc file
    """

    # https://regex101.com/
    patt_match = re.compile(r'\[([a-zA-Z]+)\s?=\s?([0-9]+)\]')

    capture = False
    block = []
    items = []
    kind = None
    sequence = 0

    f = open(fn, 'r')
    for line in f:
        line = line.strip()
        if not line:
            continue

        m = re.match(patt_match, line)

        if m:
            if block:
                items.append(block2dict(block, kind=kind, sequence=sequence))

            # prep for next block
            kind = m.groups()[0]
            sequence = int(m.groups()[1])

            block = []
            capture = True
        elif capture:
            block.append(line)
        else:
            print(line)

    items.append(block2dict(block, kind=kind, sequence=sequence))

    if only_kind:
        only_kind = only_kind.lower()
        items = [item for item in items if item['kind'].lower() == only_kind]

    return items


if __name__ == '__main__':
    fn = 'C:/s/work_2019-06-26/navs2.nav'
    items = read_nav_file(fn)

    from IPython import embed
    embed()



import numpy as np
import pytest

from pyserialem.utils import bin_ndarray


def test_binning():
    x = 32
    y = 32
    x2 = int(x / 2)
    y2 = int(y / 2)
    a = np.arange(x * y).reshape(x, y)

    b = bin_ndarray(a, new_shape=(x2, y2))
    assert b.shape == (x2, y2)

    b = bin_ndarray(a, new_shape=(x, y))
    assert b.shape == (x, y)

    b = bin_ndarray(a, binning=2)
    assert b.shape == (x2, y2)

    with pytest.raises(ValueError):
        b = bin_ndarray(a, binning=1.32)


def test_value():
    m = np.arange(0, 100, 1).reshape((10, 10))
    n = bin_ndarray(m, new_shape=(5, 5), operation='sum')
    m = np.array([
        [22, 30, 38, 46, 54],
        [102, 110, 118, 126, 134],
        [182, 190, 198, 206, 214],
        [262, 270, 278, 286, 294],
        [342, 350, 358, 366, 374],
    ])
    assert np.allclose(n, m)



# DO NOT EDIT THIS FILE!
# This file has been autogenerated by dephell <3
# https://github.com/dephell/dephell

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


import os.path

readme = ''
here = os.path.abspath(os.path.dirname(__file__))
readme_path = os.path.join(here, 'README.rst')
if os.path.exists(readme_path):
    with open(readme_path, 'rb') as stream:
        readme = stream.read().decode('utf8')


setup(
    long_description=readme,
    name='pyserialem',
    version='0.3.2',
    description='Python module to read/write SerialEM .nav files.',
    python_requires='>=3.6.1',
    project_urls={
        'documentation': 'http://github.com/stefsmeets/pyserialem',
        'homepage': 'http://github.com/stefsmeets/pyserialem',
        'repository': 'http://github.com/stefsmeets/pyserialem'},
    author='Stef Smeets',
    author_email='s.smeets@esciencecenter.nl',
    license='BSD-3-clause',
    keywords='serialem electron-microscopy navigator',
    classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Topic :: Software Development :: Libraries',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9'],
    packages=['pyserialem'],
    package_dir={
        '': '.'},
    package_data={},
    install_requires=[
        'lmfit>=1.0.1',
        'matplotlib>=3.1.2',
        'mrcfile>=1.1.2',
        'numpy>=1.17.3',
        'scikit-image>=0.17.2',
        'scipy>=1.5.0',
        'tqdm>=4.46.1'],
    extras_require={
        'dev': [
            'check-manifest',
            'pre-commit',
            'pytest==5.*,>=5.4.1',
            'pytest-cov==2.*,>=2.8.1']},
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyserialem']

package_data = \
{'': ['*']}

install_requires = \
['lmfit>=1.0.1',
 'matplotlib>=3.1.2',
 'mrcfile>=1.1.2',
 'numpy>=1.17.3',
 'scikit-image>=0.17.2',
 'scipy>=1.5.0',
 'tqdm>=4.46.1']

setup_kwargs = {
    'name': 'pyserialem',
    'version': '0.3.2',
    'description': 'Python module to read/write SerialEM .nav files.',
    'long_description': "[![GitHub Workflow Status](https://img.shields.io/github/workflow/status/instamatic-dev/pyserialem/build)](https://github.com/instamatic-dev/pyserialem/actions)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pyserialem)](https://pypi.org/project/pyserialem/)\n[![PyPI](https://img.shields.io/pypi/v/pyserialem.svg?style=flat)](https://pypi.org/project/pyserialem/)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/pyserialem)](https://pypi.org/project/pyserialem/)\n\n# PySerialEM\n\nA small Python library to read and write [SerialEM][serialem] navigator files (`.nav`), and process grid maps.\n\nInstall using `pip install pyserialem`.\n\n# Usage\n\nThe main use of `pyserialem` is to manipulate coordinates in a `.nav` file written by SerialEM ([specification][serialem_nav]).\n\nReading a `.nav` file:\n\n```python\nimport pyserialem\nfrom pathlib import Path\n\np = Path('C:/path/to/data/') / 'nav.nav'\nitems = pyserialem.read_nav_file(p)  # list\n```\nYou can set the `acquire_only` toggle to return only the items with the `Acquire` tag set:\n\n```python\nitems = pyserialem.read_nav_file(p, acquire_only=True)  # list\n```\n\nThis returns a `list` of `MapItem` and `NavItem`. A `MapItem` is associated with an image in the corresponding `.mrc` file, and a `NavItem` is a marker or point on that image.\n\n```python\nmap_items = [item.kind == 'Map' for item in items]\nnav_items = [item.kind == 'Marker' for item in items]\n```\n\nAll of the tags associated with the `MapItem` or `NavItem` can be accessed as an attribute using the same name as in the `.nav` file, i.e. with the key defined [here][serialem_nav]. This is also how the values should be updated:\n\n```python\nnav_item = nav_items[0]\nstage_position = nav_item.StageXYZ  # tuple\nmap_item.StageXYZ = (100, 200, 0)   # overwrite values\n```\n\nAlternatively, the stage position can be accessed directly through:\n\n```python\nx = map_item.stage_x\ny = map_item.stage_y\nz = map_item.stage_z\nxy = map_item.stage_xy\n```\n\n## Map items\n\nA `MapItem` has all the functions of a `NavItem`, and then some. Each `MapItem` can have a list of markers associated with it:\n\n```python\nmap_item = map_items[0]\nmarkers = map_item.markers  # list\n```\n\nTo visualize them, call:\n\n```python\nmap_item.plot()\n```\n\nTo just load the image associated with the `MapItem`:\n\n```python\nimg = map_item.load()  # np.array\n```\n\nThey can be extracted as a dictionary:\n\n```python\nd = map_item.to_dict()  # dict\n```\n\n...and restored:\n\n```python\nnew_map_item = pysem.from_dict(d, tag='new_mapitem')\n```\n\nThis is also the easiest way of constructing a new `MapItem`, because some keys can be autogenerated. Otherwise, all the required keys have to be specified to the `MapItem` constructor. The `tag` specifies the name of the item when displayed in `SerialEM`. If omitted, one will be generated.\n\nIt is easy to add new markers to a `MapItem`. As a pixel coordinate (i.e. from segmentation) is the default. `PySerialEM` calculates the corresponding stage position. The `acquire` argument sets the acquire flag (default=`True`):\n\n```python\npixel_position = (0, 0)\nnew_nav_item = map_item.add_marker(\n    pixel_position,\n    tag='pixel_item',\n    acquire=True)  # NavItem\n```\n\nYou can also add a marker as a stage coordinate (although this is a bit more tricky to calculate the corresponding pixel coordinate):\n\n```python\nstage_positionion = (1000, 1000)\nnew_nav_item = map_item.add_marker(\n    pixel_position,\n    kind='stage',\n    tag='stage_item',\n    acquire=False)  # NavItem\n```\n\nTo add many markers:\n\n```python\npixel_coordinates = ((0, 0), (100, 100), (200, 200))\nnav_item_group = map_item.add_marker_group(pixel_coordinates)  # tuple\n```\n\nSpecify `replace=True` to replace the current list of markers associated with the `MapItem`.\n\nIf the `MapItem` has a set of markers associated with it `map_item.markers`, the coordinates be retrieved as followed:\n\n```python\nmap_item.markers_as_pixel_coordinates()  # np.array (Nx2)\nmap_item.markers_as_stage_coordinates()  # np.array (Nx2)\n```\n\nTo just convert between stage and pixel coordinates:\n\n```python\npixel_coord = (1024, 1024)\nstage_coord = map_item.pixel_to_stagecoords(pixel_coord)  # tuple\nnew_pixel_coord = map_item.stage_to_pixelcoords(stagecoord)  # tuple\nassert new_pixel_coord == pixel_coord\n```\n\nTo write a new file:\n\n```python\npyserialem.write_nav_file('out.nav', map_item, *nav_item_group)\n```\n\nNote the `*`. This function captures arguments in a list (`*args`, so they must be unpacked when supplied.\n\n## Stitching\n\nA basic stitching algorithm is available to get an overview of the location of all map items:\n\n```python\nmap_items = [item for item in items if item.kind == 'Map']\npyserialem.stitch_map_items(map_items)\n```\n\nFor more advanced stitching and montaging, use the `pyserialem.montage` module. A [demo notebook](demos/montage_processing_serialem.ipynb) is available to demonstrate its usage.\n\n## Mdoc files\n\nThere is also a simple function to read `.mdoc` files ([link][serialem_nav]). This returns a list of python objects where each key can be accessed as an attribute.\n\n```python\np = Path('C:/path/to/data') / 'gm.mrc.mdoc'\nmdoc = pyserialem.read_mdoc_file(p)\n```\n\n[src]: https://github.com/stefsmeets/pyserialem\n[serialem]: https://bio3d.colorado.edu/SerialEM/\n[serialem_nav]: https://bio3d.colorado.edu/SerialEM/hlp/html/about_formats.htm\n",
    'author': 'Stef Smeets',
    'author_email': 's.smeets@tudelft.nl',
    'maintainer': 'Stef Smeets',
    'maintainer_email': 's.smeets@tudelft.nl',
    'url': 'http://github.com/stefsmeets/pyserialem',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)



import matplotlib.pyplot as plt
import numpy as np


def closest_distance(node, nodes: list) -> np.array:
    """Get shortest between a node and a list of nodes (that includes the given
    node)"""
    nodes = np.asarray(nodes)
    dist_2 = np.linalg.norm(nodes - node, axis=1)
    return np.sort(dist_2)[1]


def filter_nav_items_by_proximity(items, min_sep: float = 5.0) -> np.array:
    """Filter navigator items (markers) if they are within `min_sep` micrometer
    of another one."""
    ret = []
    stagecoords = np.array([item.stage_xy for item in items])
    for i, coord in enumerate(stagecoords):
        try:
            min_dist = closest_distance(coord, stagecoords)
        except IndexError:
            min_dist = np.inf

        if min_dist > min_sep:
            ret.append(items[i])

    return ret


def calc_total_dist(coords: list, route: list = None) -> float:
    """Calculate the total distance over a list of (x,y) coordinates.

    route, list[int]
        List of integers with the same length as `coords` that specifies
        the order of the items

    Returns:
        total_dist, float
            Total distance for the given path
    """
    if route is None:
        diffs = np.diff(coords, axis=0)
    else:
        diffs = np.diff(coords[route], axis=0)

    # TODO: add backlash in diffs/total_dist calculation
    total_dist = np.linalg.norm(diffs, axis=1).sum()

    return total_dist


def two_opt_swap(r: list, i: int, k: int) -> list:
    """Reverses items `i`:`k` in the list `r`
    https://en.wikipedia.org/wiki/2-opt."""
    out = r.copy()
    out[i:k + 1] = out[k:i - 1:-1]
    return out


def two_opt(coords: list, threshold: float, verbose: bool = False) -> list:
    """Implementation of the two_opt algorithm for finding the shortest path
    through a list of coordinates (x, y)
    https://en.wikipedia.org/wiki/2-opt."""
    route = np.arange(len(coords))
    improvement = 1
    initial_distance = best_distance = calc_total_dist(coords, route)
    while improvement > threshold:
        previous_best_distance = best_distance
        for i in range(1, len(route) - 2):
            for k in range(i + 1, len(route)):
                new_route = two_opt_swap(route, i, k)
                new_distance = calc_total_dist(coords, new_route)
                if new_distance < best_distance:
                    route = new_route
                    best_distance = new_distance
        improvement = 1 - best_distance / previous_best_distance
    if verbose:
        diff = initial_distance - best_distance
        perc = diff / initial_distance
        print(f'Optimized path for {len(coords)} items from {initial_distance/1000:.1f} to {best_distance/1000:.1f} μm (-{perc:.2%})')

    return route


def sort_nav_items_by_shortest_path(items: list,
                                    first: int = 0,
                                    threshold: float = 0.1,
                                    plot: bool = False,
                                    ) -> list:
    """Find shortest route based on stage coordinates (.stage_xy)

    Parameters
    ----------
    items : list
        List of navigation items with stage coordinates (item.stage_xy)
    first : int or tuple
        If first is an integer, it is the index of the object to start from, must be 0 <= x < len(items)
        If it is a tuple, it should contain the stage coordinates, e.g. the last known coordinate, the function will figure out
        the closest coordinate to start from.
    threshold : float
        Number between 0.0 and 1.0 that determines when convergence is reached. If the improvement is smaller than the threshold, the algorithm will accept convergence.
    plot : bool
        Plot the resulting Path

    Returns
    -------
        List of navigation items sorted to minimize total path distance
    """
    try:
        coords = np.array([item.stage_xy for item in items])
        is_nav = True
    except AttributeError:
        coords = items
        is_nav = False

    if not isinstance(first, int):  # assume it is a coordinate
        first = np.array(first)
        first = np.argmin(np.linalg.norm(coords - first, axis=1))
        print(f'First = {first}')

    if first > 0:
        coords = np.concatenate((coords[first:first + 1], coords[0:first], coords[first + 1:]))

    route = two_opt(coords, threshold, verbose=True)

    if plot:
        fig, (ax0, ax1) = plt.subplots(ncols=2, figsize=(12, 6))
        new_coords = coords[route]
        ax0.set_title(f'Before, total distance: {calc_total_dist(coords)/1000:.3g} μm')
        ax0.plot(coords[:, 0] / 1000, coords[:, 1] / 1000, 'r--', marker='.', lw=1.0)
        ax0.scatter(coords[0, 0] / 1000, coords[0, 1] / 1000, color='red', s=100)
        ax0.axis('equal')
        ax1.set_title(f'After, total distance: {calc_total_dist(new_coords)/1000:.3g} μm')
        ax1.plot(new_coords[:, 0] / 1000, new_coords[:, 1] / 1000, 'r--', marker='.', lw=1.0)
        ax1.scatter(new_coords[0, 0] / 1000, new_coords[0, 1] / 1000, color='red', s=50)
        ax1.axis('equal')
        plt.show()

    if is_nav:
        return [items[i] for i in route]
    else:
        return items[route]
